"""MIT License

Copyright (c) 2019-2022 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import argparse
import asyncio
import pathlib
import sys
import typing

import aiohttp


parser = argparse.ArgumentParser()
parser.add_argument('--java', action='store_true')

args = parser.parse_args()
get_java = args.java


RELEASES = 'https://api.github.com/repos/freyacodes/Lavalink/releases/latest'
JAVA = 'https://download.oracle.com/java/17/archive/jdk-17.0.3.1_windows-x64_bin.exe'


if get_java and sys.platform != 'win32':
    raise RuntimeError('Downloading and installing Java is only supported on Windows.')

if get_java:
    import ctypes
    import os

    try:
        is_admin = os.getuid() == 0
    except AttributeError:
        is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0

    if not is_admin:
        raise RuntimeError('This script requires administrator privileges to install Java.\n'
                           'Please restart the script in Command Prompt or Powershell as an administrator.')


_application_yml = """
server: # REST and WS server
  port: {port}
  address: {address}
lavalink:
  server:
    password: "{password}"
    sources:
      youtube: true
      bandcamp: true
      soundcloud: true
      twitch: true
      vimeo: true
      http: true
      local: false
    bufferDurationMs: 400 # The duration of the NAS buffer. Higher values fare better against longer GC pauses
    frameBufferDurationMs: 5000 # How many milliseconds of audio to keep buffered
    youtubePlaylistLoadLimit: 6 # Number of pages at 100 each
    playerUpdateInterval: 5 # How frequently to send player updates to clients, in seconds
    youtubeSearchEnabled: true
    soundcloudSearchEnabled: true
    gc-warnings: true
    #ratelimit:
      #ipBlocks: ["1.0.0.0/8", "..."] # list of ip blocks
      #excludedIps: ["...", "..."] # ips which should be explicit excluded from usage by lavalink
      #strategy: "RotateOnBan" # RotateOnBan | LoadBalance | NanoSwitch | RotatingNanoSwitch
      #searchTriggersFail: true # Whether a search 429 should trigger marking the ip as failing
      #retryLimit: -1 # -1 = use default lavaplayer value | 0 = infinity | >0 = retry will happen this numbers times

metrics:
  prometheus:
    enabled: false
    endpoint: /metrics

sentry:
  dsn: ""
  environment: ""
#  tags:
#    some_key: some_value
#    another_key: another_value

logging:
  file:
    max-history: 30
    max-size: 1GB
  path: ./logs/

  level:
    root: INFO
    lavalink: INFO
"""


async def download(location: pathlib.Path, *, url: str) -> None:
    lavalink = None

    async with aiohttp.ClientSession() as session:
        async with session.get(url=url) as head_resp:
            if url == RELEASES:

                data = await head_resp.json()
                assets = data['assets']

                for asset in assets:
                    if asset['name'] == 'Lavalink.jar':

                        length = asset['size']
                        lavalink = asset['browser_download_url']
                        chunks = round(length / 1024)
                        break
            else:
                length = head_resp.content_length
                chunks = round(length / 1024)

        if lavalink:
            file = 'Lavalink.jar'
            url = lavalink
        else:
            file = 'java17.exe'

        async with session.get(url=url) as resp:
            with open(f'{location.absolute()}/{file}', 'wb') as fp:
                count = 1

                async for chunk in resp.content.iter_chunked(1024):
                    x = int(40 * count / chunks)

                    print(f"\rDownloading {file}: [{u'█' * x}{('.' * (40 - x))}] "
                          f"{count / 1024:.2f}/{chunks / 1024:.2f} MB",
                          end='',
                          file=sys.stdout,
                          flush=True)

                    fp.write(chunk)
                    count += 1

        print("\n", flush=True, file=sys.stdout)


def parse_input(value: str, type_: typing.Any) -> typing.Union[str, int, bool, None]:
    value = str(value)

    quits = ('q', 'quit', 'exit')
    if value.lower() in quits:
        print('Exiting installer...')
        sys.exit(0)

    if type_ is bool:

        bools = {'y': True, 'yes': True, 'true': True, 'n': False, 'no': False, 'false': False}
        result = bools.get(value.lower(), None)

        return result

    try:
        result = type_(value.lower())
    except ValueError:
        return None

    return result


async def main():
    cwd = pathlib.Path.cwd()

    print('\nEnter Q at anytime to quit the installer...\n')

    while True:
        while True:

            port = parse_input(input('Please enter the port number for Lavalink (Enter to use default: 2333): ')
                               or 2333, int)

            if not port:
                print('Invalid port specified. Please enter a number for your port. Enter Q to quit.')
                continue

            break

        while True:

            address = parse_input(input('Please enter the address to star Lavalink (Enter to use default: 0.0.0.0): ')
                                  or '0.0.0.0', str)

            if not address:
                print('Invalid address specified. Please enter a valid binding IP address. Enter Q to quit.')
                continue

            break

        while True:

            password = parse_input(input('Please enter a password for Lavalink (Enter to use default:'
                                         ' "youshallnotpass"): ')
                                   or 'youshallnotpass', str)

            if not password:
                print('Invalid password specified. Please enter a valid password. Enter Q to quit.')
                continue

            break

        while True:

            location = parse_input(input('Please enter a install directory. '
                                         'An attempt to create the directory will be made if it does not not exist. '
                                         f'(Enter to use the current directory: {cwd}): ')
                                   or cwd, str)

            if not location:
                print('Invalid location specified. Please enter a valid directory. Enter Q to quit.')
                continue

            location = pathlib.Path(location)

            break

        while True:
            confirmed_ = parse_input(input('\nPlease confirm the following information:\n\n'
                                           f'Port: {port}\nAddress:'
                                           f' {address}\nPassword:'
                                           f' "{password}"\nLocation:'
                                           f' {location.absolute()}\n\n'
                                           f'(Y: Confirm, N: Re-Enter, Q: Quit): '),
                                     bool)

            if confirmed_ is None:
                print('Invalid response...')
                continue

            break

        if confirmed_ is True:
            break

    if not location.exists():
        try:
            location.mkdir(parents=True, exist_ok=False)
        except Exception as e:
            print(f'Could not make directory: "{location.absolute()}". {e}\nExiting the installer...')

    await download(location=location, url=RELEASES)
    print('Lavalink successfully downloaded...')

    with open(f'{location.absolute()}/application.yml', 'w') as fp:
        fp.write(_application_yml.format(port=port, address=address, password=password))

    if get_java:
        await download(location=location, url=JAVA)

        proc = await asyncio.subprocess.create_subprocess_exec(f'{location.absolute()}/java17.exe')
        await proc.wait()

    print(f'\nSuccess!\n\nDownload Location: {location.absolute()}\n\nInstructions:\n'
          f'- Open a Powershell or Command Prompt\n'
          f'- cd {location.absolute()}\n'
          f'- java -jar Lavalink.jar\n\n'
          f'To connect use the following details:\n'
          f'Port     : {port}\n'
          f'Password : "{password}"\n\n'
          f'You may change these values by editing the application.yml\n')


if __name__ == '__main__':
    asyncio.run(main())



"""MIT License

Copyright (c) 2019-2021 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from .enums import ErrorSeverity
from discord.enums import try_enum


__all__ = (
    "WavelinkError",
    "AuthorizationFailure",
    "LavalinkException",
    "LoadTrackError",
    "BuildTrackError",
    "NodeOccupied",
    "InvalidIDProvided",
    "ZeroConnectedNodes",
    "NoMatchingNode",
    "QueueException",
    "QueueFull",
    "QueueEmpty",
)


class WavelinkError(Exception):
    """Base WaveLink Exception"""


class AuthorizationFailure(WavelinkError):
    """Exception raised when an invalid password is provided toa node."""


class LavalinkException(WavelinkError):
    """Exception raised when an error occurs talking to Lavalink."""


class LoadTrackError(LavalinkException):
    """Exception raised when an error occurred when loading a track."""

    def __init__(self, data):
        exception = data["exception"]
        self.severity: ErrorSeverity
        super().__init__(exception["message"])


class BuildTrackError(LavalinkException):
    """Exception raised when a track is failed to be decoded and re-built."""

    def __init__(self, data):
        super().__init__(data["error"])


class NodeOccupied(WavelinkError):
    """Exception raised when node identifiers conflict."""


class InvalidIDProvided(WavelinkError):
    """Exception raised when an invalid ID is passed somewhere in Wavelink."""


class ZeroConnectedNodes(WavelinkError):
    """Exception raised when an operation is attempted with nodes, when there are None connected."""


class NoMatchingNode(WavelinkError):
    """Exception raised when a Node is attempted to be retrieved with a incorrect identifier."""


class QueueException(WavelinkError):
    """Base WaveLink Queue exception."""

    pass


class QueueFull(QueueException):
    """Exception raised when attempting to add to a full Queue."""

    pass


class QueueEmpty(QueueException):
    """Exception raised when attempting to retrieve from an empty Queue."""

    pass



"""MIT License

Copyright (c) 2019-2021 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    ClassVar,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

import yarl
from discord.ext import commands

from .abc import *
from .pool import Node, NodePool
from .utils import MISSING

if TYPE_CHECKING:
    from .ext import spotify


__all__ = (
    "Track",
    "SearchableTrack",
    "YouTubeTrack",
    "YouTubeMusicTrack",
    "SoundCloudTrack",
    "YouTubePlaylist",
    "PartialTrack",
    "LocalTrack"
)

ST = TypeVar("ST", bound="SearchableTrack")


class Track(Playable):
    """A Lavalink track object.

    Attributes
    ------------
    id: str
        The Base64 Track ID, can be used to rebuild track objects.
    info: Dict[str, Any]
        The raw track info.
    title: str
        The track title.
    identifier: Optional[str]
        The tracks identifier. could be None depending on track type.
    length:
        The duration of the track in seconds.
    duration:
        Alias to ``length``.
    uri: Optional[str]
        The tracks URI. Could be None.
    author: Optional[str]
        The author of the track. Could be None
    """

    def __init__(self, id: str, info: dict):
        super().__init__(id, info)
        self.title: str = info.get('title')
        self.identifier: Optional[str] = info.get("identifier")
        self.uri: Optional[str] = info.get("uri")
        self.author: Optional[str] = info.get("author")

        self._stream: bool = info.get("isStream")  # type: ignore
        self._dead: bool = False

    def __str__(self) -> str:
        return self.title

    def is_stream(self) -> bool:
        """Indicates whether the track is a stream or not."""
        return self._stream


class SearchableTrack(Track, Searchable):

    _search_type: ClassVar[str]

    @overload
    @classmethod
    async def search(
            cls: Type[ST],
            query: str,
            *,
            node: Node = ...,
            type: spotify.SpotifySearchType = ...,
            return_first: Literal[False] = ...,
    ) -> List[ST]:
        ...

    @overload
    @classmethod
    async def search(
            cls: Type[ST],
            query: str,
            *,
            node: Node = ...,
            type: spotify.SpotifySearchType = ...,
            return_first: Literal[True] = ...,
    ) -> Optional[ST]:
        ...

    @overload
    @classmethod
    async def search(
        cls: Type[ST],
        query: str,
        *,
        node: Node = ...,
        return_first: Literal[False] = ...,
    ) -> List[ST]:
        ...

    @overload
    @classmethod
    async def search(
        cls: Type[ST],
        query: str,
        *,
        node: Node = ...,
        return_first: Literal[True] = ...,
    ) -> Optional[ST]:
        ...

    @classmethod
    async def search(
        cls: Type[ST],
            query: str,
            *,
            type: spotify.SpotifySearchType = None,
            node: Node = MISSING,
            return_first: bool = False
    ) -> Union[Optional[ST], List[ST]]:
        """|coro|

        Search for tracks with the given query.

        Parameters
        ----------
        query: str
            The song to search for.
        spotify_type: Optional[:class:`spotify.SpotifySearchType`]
            An optional enum value to use when searching with Spotify.
        node: Optional[:class:`wavelink.Node`]
            An optional Node to use to make the search with.
        return_first: Optional[bool]
            An optional bool which when set to True will return only the first track found. Defaults to False.
            Use this as True, when searching with LocalTrack.

        Returns
        -------
        Union[Optional[Track], List[Track]]
        """
        if node is MISSING:
            node = NodePool.get_node()

        check = yarl.URL(query)

        if str(check.host) == 'youtube.com' or str(check.host) == 'www.youtube.com' and check.query.get("list") or \
                cls._search_type == 'ytpl':

            tracks = await node.get_playlist(cls=YouTubePlaylist, identifier=query)
        elif cls._search_type == 'local':
            tracks = await node.get_tracks(cls, query)
        else:
            tracks = await node.get_tracks(cls, f"{cls._search_type}:{query}")

        if return_first:
            return tracks[0]

        return tracks

    @classmethod
    async def convert(cls: Type[ST], ctx: commands.Context, argument: str) -> ST:
        """Converter which searches for and returns the first track.

        Used as a type hint in a discord.py command.
        """
        if argument.startswith('local:'):
            argument.replace('local:', '')

        results = await cls.search(argument)

        if not results:
            raise commands.BadArgument("Could not find any songs matching that query.")

        if issubclass(cls, YouTubePlaylist):
            return results  # type: ignore

        return results[0]


class YouTubeTrack(SearchableTrack):
    """A track created using a search to YouTube."""

    _search_type: ClassVar[str] = "ytsearch"

    @property
    def thumbnail(self) -> str:
        """The URL to the thumbnail of this video."""
        return f"https://img.youtube.com/vi/{self.identifier}/maxresdefault.jpg"

    thumb = thumbnail


class YouTubeMusicTrack(YouTubeTrack):
    """A track created using a search to YouTube Music."""

    _search_type: ClassVar[str] = "ytmsearch"


class SoundCloudTrack(SearchableTrack):
    """A track created using a search to SoundCloud."""

    _search_type: ClassVar[str] = "scsearch"


class YouTubePlaylist(SearchableTrack, Playlist):
    """Represents a Lavalink YouTube playlist object.

    Attributes
    ----------
    name: str
        The name of the playlist.
    tracks: :class:`YouTubeTrack`
        The list of :class:`YouTubeTrack` in the playlist.
    selected_track: Optional[int]
        The selected video in the playlist. This could be ``None``.
    """

    _search_type: ClassVar[str] = "ytpl"

    def __init__(self, data: dict):
        self.tracks: List[YouTubeTrack] = []
        self.name: str = data["playlistInfo"]["name"]

        self.selected_track: Optional[int] = data["playlistInfo"].get("selectedTrack")
        if self.selected_track is not None:
            self.selected_track = int(self.selected_track)

        for track_data in data["tracks"]:
            track = YouTubeTrack(track_data["track"], track_data["info"])
            self.tracks.append(track)

    def __str__(self) -> str:
        return self.name


class LocalTrack(SearchableTrack):
    """Represents a Lavalinkl Local Track Object."""

    _search_type: ClassVar[str] = 'local'


class PartialTrack(Searchable, Playable):
    """A PartialTrack object that searches for the given query at playtime.

    Parameters
    ----------
    query: str
        The query to search for at playtime.
    node: Optional[:class:`wavelink.Node`]
        An optional node to use when searching. Defaults to the best node.
    cls: Optional[:class:`SearchableTrack`]
        An optional Non-Partial Track object to use when searching.


    .. warning::

        This object will only search for the given query at playtime.
        Full track information will be missing until it has been searched.
    """

    def __init__(self,
                 *,
                 query: str,
                 node: Optional[Node] = MISSING,
                 cls: Optional[SearchableTrack] = YouTubeTrack):
        self.query = query
        self.title = query
        self._node = node
        self._cls = cls

        if not issubclass(cls, SearchableTrack):
            raise TypeError(f"cls parameter must be of type {SearchableTrack!r} not {cls!r}")

    async def _search(self):
        node = self._node
        if node is MISSING:
            node = NodePool.get_node()

        tracks = await self._cls.search(query=self.query, node=node)

        return tracks[0]

    async def search(self):
        raise NotImplementedError



"""MIT License

Copyright (c) 2019-2021 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import json
import logging
import os
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

import aiohttp
import discord
from discord.enums import try_enum

from . import abc
from .enums import *
from .errors import *
from .stats import Stats
from .utils import MISSING
from .websocket import Websocket

if TYPE_CHECKING:
    from .player import Player
    from .ext import spotify


__all__ = (
    "Node",
    "NodePool",
)


PT = TypeVar("PT", bound=abc.Playable)
PLT = TypeVar("PLT", bound=abc.Playlist)


logger: logging.Logger = logging.getLogger(__name__)


class Node:
    """WaveLink Node object.

    Attributes
    ----------
    bot: :class:`discord.Client`
        The discord.py Bot object.


    .. warning::
        This class should not be created manually. Please use :meth:`NodePool.create_node()` instead.
    """

    def __init__(
        self,
        bot: discord.Client,
        host: str,
        port: int,
        password: str,
        https: bool,
        heartbeat: float,
        region: Optional[discord.VoiceRegion],
        spotify: Optional[spotify.SpotifyClient],
        identifier: str,
        dumps: Callable[[Any], str],
        resume_key: Optional[str],
    ):
        self.bot: discord.Client = bot
        self._host: str = host
        self._port: int = port
        self._password: str = password
        self._https: bool = https
        self._heartbeat: float = heartbeat
        self._region: Optional[discord.VoiceRegion] = region
        self._spotify = spotify
        self._identifier: str = identifier

        self._players: List[Player] = []

        self._dumps: Callable[[Any], str] = dumps
        self._websocket: Websocket = MISSING

        self.stats: Optional[Stats] = None

        self.resume_key = resume_key or str(os.urandom(8).hex())

    def __repr__(self) -> str:
        return f"<WaveLink Node: <{self.identifier}>, Region: <{self.region}>, Players: <{len(self._players)}>>"

    @property
    def host(self) -> str:
        """The host this node is connected to."""
        return self._host

    @property
    def port(self) -> int:
        """The port this node is connected to."""
        return self._port

    @property
    def region(self) -> Optional[discord.VoiceRegion]:
        """The voice region of the Node."""
        return self._region

    @property
    def identifier(self) -> str:
        """The Nodes unique identifier."""
        return self._identifier

    @property
    def players(self) -> List[Player]:
        """A list of currently connected Players."""
        return self._players

    @property
    def penalty(self) -> float:
        """The load-balancing penalty for this node."""
        if self.stats is None:
            return 9e30

        return self.stats.penalty.total

    def is_connected(self) -> bool:
        """Bool indicating whether or not this Node is currently connected to Lavalink."""
        if self._websocket is MISSING:
            return False

        return self._websocket.is_connected()

    async def _connect(self) -> None:
        self._websocket = Websocket(node=self)

        await self._websocket.connect()

    async def _get_data(
        self, endpoint: str, params: dict
    ) -> Tuple[Dict[str, Any], aiohttp.ClientResponse]:
        headers = {"Authorization": self._password}
        async with self._websocket.session.get(
            f"{self._websocket.host}/{endpoint}", headers=headers, params=params
        ) as resp:
            data = await resp.json()

        return data, resp

    async def get_tracks(self, cls: Type[PT], query: str) -> List[PT]:
        """|coro|

        Search for and return a list of :class:`abc.Playable` for a given query.

        Parameters
        ----------
        cls: Type[:class:`abc.Playable`]
            The type of which track should be returned, this must subclass :class:`abc.Playable`.
        query: str
            A query to use to search for tracks.

        Returns
        -------
        List[:class:`abc.Playable`]:
            A list of wavelink track objects.

        Raises
        ------
        LoadTrackError
            Loading the track failed.
        LavalinkException
            An unspecified error occurred when loading the track.
        """
        data, resp = await self._get_data("loadtracks", {"identifier": query})

        if resp.status != 200:
            raise LavalinkException("Invalid response from Lavalink server.")

        load_type = try_enum(LoadType, data.get("loadType"))

        if load_type is LoadType.load_failed:
            raise LoadTrackError(data)

        if load_type is LoadType.no_matches:
            return []

        if load_type is LoadType.track_loaded:
            track_data = data["tracks"][0]
            return [cls(track_data["track"], track_data["info"])]

        if load_type is not LoadType.search_result:
            raise LavalinkException("Track failed to load.")

        return [
            cls(track_data["track"], track_data["info"])
            for track_data in data["tracks"]
        ]

    async def get_playlist(self, cls: Type[PLT], identifier: str) -> Optional[PLT]:
        """|coro|

        Search for and return a :class:`abc.Playlist` given an identifier.

        Parameters
        ----------
        cls: Type[:class:`abc.Playlist`]
            The type of which playlist should be returned, this must subclass :class:`abc.Playlist`.
        identifier: str
            The playlist's identifier. This may be a YouTube playlist URL for example.

        Returns
        -------
        Optional[:class:`abc.Playlist`]:
            The related wavelink track object or ``None`` if none was found.

        Raises
        ------
        LoadTrackError
            Loading the playlist failed.
        LavalinkException
            An unspecified error occurred when loading the playlist.
        """
        data, resp = await self._get_data("loadtracks", {"identifier": identifier})

        if resp.status != 200:
            raise LavalinkException("Invalid response from Lavalink server.")

        load_type = try_enum(LoadType, data.get("loadType"))

        if load_type is LoadType.load_failed:
            raise LoadTrackError(data)

        if load_type is LoadType.no_matches:
            return None

        if load_type is not LoadType.playlist_loaded:
            raise LavalinkException("Track failed to load.")

        return cls(data)

    async def build_track(self, cls: Type[PT], identifier: str) -> PT:
        """|coro|

        Build a track object with a valid track identifier.

        Parameters
        ------------
        cls: Type[:class:`abc.Playable`]
            The type of which track should be returned, this must subclass :class:`abc.Playable`.
        identifier: str
            The tracks unique Base64 encoded identifier. This is usually retrieved from various lavalink events.

        Returns
        ---------
        :class:`abc.Playable`
            The track built from a Base64 identifier.

        Raises
        --------
        BuildTrackError
            Decoding and building the track failed.
        """
        data, resp = await self._get_data("decodetrack", {"track": identifier})

        if resp.status != 200:
            raise BuildTrackError(data)

        return cls(identifier, data)

    def get_player(self, guild: discord.Guild) -> Optional[Player]:
        """Returns a :class:`Player` object playing in a specific :class:`discord.Guild`.

        Parameters
        ----------
        guild: :class:`discord.Guild`
            The Guild the player is in.

        Returns
        -------
        Optional[:class:`Player`]
        """
        for player in self.players:
            if player.guild == guild:
                return player

        return None

    async def disconnect(self, *, force: bool = False) -> None:
        """Disconnect this Node and remove it from the NodePool.

        This is a graceful shutdown of the node.
        """
        for player in self.players:
            await player.disconnect(force=force)

        await self.cleanup()

    async def cleanup(self) -> None:
        try:
            self._websocket.listener.cancel()
        except Exception:
            pass

        try:
            await self._websocket.session.close()
        except Exception:
            pass

        del NodePool._nodes[self._identifier]


class NodePool:
    """Wavelink NodePool class.

    This class holds all the Node objects created with :meth:`create_node()`.
    """

    _nodes: ClassVar[Dict[str, Node]] = {}

    @property
    def nodes(self) -> Dict[str, Node]:
        """A mapping of created Node objects."""
        return self._nodes

    @classmethod
    async def create_node(
        cls,
        *,
        bot: discord.Client,
        host: str,
        port: int,
        password: str,
        https: bool = False,
        heartbeat: float = 30,
        region: Optional[discord.VoiceRegion] = None,
        spotify_client: Optional[spotify.SpotifyClient] = None,
        identifier: str = MISSING,
        dumps: Callable[[Any], str] = json.dumps,
        resume_key: Optional[str] = None,
    ) -> Node:

        """|coro|

        Classmethod that creates a :class:`Node` object and stores it for use with WaveLink.

        Parameters
        ----------
        bot: Union[:class:`discord.Client`]
            The discord.py Bot or Client class.
        host: :class:`str`
            The lavalink host address.
        port: :class:`int`
            The lavalink port.
        password: :class:`str`
            The lavalink password for authentication.
        https: :class:`bool`
            Connect to lavalink over https. Defaults to False.
        heartbeat: :class:`float`
            The heartbeat in seconds for the node. Defaults to 30 seconds.
        region: Optional[:class:`discord.VoiceRegion`]
            The discord.py VoiceRegion to assign to the node. This is useful for node region balancing.
        spotify_client: Optional[:class:`wavelink.ext.spotify.SpotifyClient`]
            An optional SpotifyClient with credentials to use when searching for spotify tracks.
        identifier: :class:`str`
            The unique identifier for this Node. By default this will be generated for you.

        Returns
        -------
        :class:`Node`
            The WaveLink Node object.
        """

        if identifier is MISSING:
            identifier = os.urandom(8).hex()

        if identifier in cls._nodes:
            raise NodeOccupied(
                f"A node with identifier <{identifier}> already exists in this pool."
            )

        node = Node(
            bot=bot,
            host=host,
            port=port,
            password=password,
            https=https,
            heartbeat=heartbeat,
            region=region,
            spotify=spotify_client,
            identifier=identifier,
            dumps=dumps,
            resume_key=resume_key
        )

        cls._nodes[identifier] = node
        await node._connect()

        return node

    @classmethod
    def get_node(
        cls, *, identifier: str = MISSING, region: discord.VoiceRegion = MISSING
    ) -> Node:
        """Retrieve a Node from the NodePool.

        Parameters
        ------------
        identifier: :class:`str`
            If provided this method will attempt to return the Node with the provided identifier.
        region: :class:`discord.VoiceRegion`
            If provided this method will attempt to find the best Node with the provided region.

        Returns
        --------
        :class:`Node`
            The WaveLink Node object.

        Raises
        --------
        :class:`ZeroConnectedNodes`
            There are no currently connected Nodes on the pool with the provided options.
        :class:`NoMatchingNode`
            No Node exists with the provided identifier.
        """
        if not cls._nodes:
            raise ZeroConnectedNodes("There are no connected Nodes on this pool.")

        if identifier is not MISSING:
            try:
                node = cls._nodes[identifier]
            except KeyError:
                raise NoMatchingNode(f"No Node with identifier <{identifier}> exists.")
            else:
                return node

        elif region is not MISSING:
            nodes = [n for n in cls._nodes.values() if n._region is region]
            if not nodes:
                raise ZeroConnectedNodes(
                    f"No Nodes for region <{region}> exist on this pool."
                )
        else:
            nodes = cls._nodes.values()

        return sorted(nodes, key=lambda n: len(n.players))[0]



"""
The MIT License (MIT)

Copyright (c) 2021 PythonistaGuild, EvieePy, Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import random
from typing import Optional


class Backoff:
    """An implementation of an Exponential Backoff.

    Parameters
    ----------
    base: int
        The base time to multiply exponentially. Defaults to 1.
    maximum_time: float
        The maximum wait time. Defaults to 30.0
    maximum_tries: Optional[int]
        The amount of times to backoff before resetting. Defaults to 5. If set to None, backoff will run indefinitely.
    """

    def __init__(self, *, base: int = 1, maximum_time: float = 30.0, maximum_tries: Optional[int] = 5):
        self._base = base
        self._maximum_time = maximum_time
        self._maximum_tries = maximum_tries
        self._retries: int = 1

        rand = random.Random()
        rand.seed()

        self._rand = rand.uniform

        self._last_wait: float = 0

    def calculate(self) -> float:
        exponent = min((self._retries ** 2), self._maximum_time)
        wait = self._rand(0, (self._base * 2) * exponent)

        if wait <= self._last_wait:
            wait = self._last_wait * 2

        self._last_wait = wait

        if wait > self._maximum_time:
            wait = self._maximum_time
            self._retries = 0
            self._last_wait = 0

        if self._maximum_tries and self._retries >= self._maximum_tries:
            self._retries = 0
            self._last_wait = 0

        self._retries += 1

        return wait



"""MIT License

Copyright (c) 2019-2021 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import contextlib
import datetime
import logging
from typing import Any, Dict, Union, Optional

import discord
from discord.channel import VoiceChannel

from . import abc
from .pool import Node, NodePool
from .queue import WaitQueue
from .tracks import PartialTrack
from .utils import MISSING
from .filters import Filter


__all__ = ("Player",)


logger: logging.Logger = logging.getLogger(__name__)


VoiceChannel = Union[
    discord.VoiceChannel, discord.StageChannel
]  # todo: VocalGuildChannel?


class Player(discord.VoiceProtocol):
    """WaveLink Player object.

    This class subclasses :class:`discord.VoiceProtocol` and such should be treated as one with additions.

    Examples
    --------

        .. code::

            @commands.command()
            async def connect(self, channel: discord.VoiceChannel):

                voice_client = await channel.connect(cls=wavelink.Player)


    .. warning::
        This class should not be created manually but can be subclassed to add additional functionality.
        You should instead use :meth:`discord.VoiceChannel.connect()` and pass the player object to the cls kwarg.
    """

    def __call__(self, client: discord.Client, channel: VoiceChannel):
        self.client: discord.Client = client
        self.channel: VoiceChannel = channel

        return self

    def __init__(
        self,
        client: discord.Client = MISSING,
        channel: VoiceChannel = MISSING,
        *,
        node: Node = MISSING,
    ):
        self.client: discord.Client = client
        self.channel: VoiceChannel = channel

        if node is MISSING:
            node = NodePool.get_node()
        self.node: Node = node
        self.node._players.append(self)

        self._voice_state: Dict[str, Any] = {}

        self.last_update: datetime.datetime = MISSING
        self.last_position: float = MISSING

        self._connected: bool = False
        self._paused: bool = False

        self._volume: int = 100
        self._source: Optional[abc.Playable] = None
        self._filter: Optional[Filter] = None

        self.queue = WaitQueue()

    @property
    def guild(self) -> discord.Guild:
        """The :class:`discord.Guild` this :class:`Player` is in."""
        return self.channel.guild

    @property
    def user(self) -> discord.ClientUser:
        """The :class:`discord.ClientUser` of the :class:`discord.Client`"""
        return self.client.user  # type: ignore

    @property
    def volume(self) -> int:
        return self._volume

    @property
    def source(self) -> Optional[abc.Playable]:
        """The currently playing audio source."""
        return self._source

    @property
    def filter(self) -> Optional[Filter]:
        return self._filter

    track = source

    @property
    def position(self) -> float:
        """The current seek position of the playing source in seconds. If nothing is playing this defaults to ``0``."""
        if not self.is_playing():
            return 0

        if self.is_paused():
            return min(self.last_position, self.source.duration)  # type: ignore

        delta = (
            datetime.datetime.now(datetime.timezone.utc) - self.last_update
        ).total_seconds()
        position = round(self.last_position + delta, 1)

        return min(position, self.source.duration)  # type: ignore

    async def update_state(self, state: Dict[str, Any]) -> None:
        state = state["state"]

        self.last_update = datetime.datetime.fromtimestamp(
            state.get("time", 0) / 1000, datetime.timezone.utc
        )
        self.last_position = round(state.get("position", 0) / 1000, 1)

    async def on_voice_server_update(self, data: Dict[str, Any]) -> None:
        self._voice_state.update({"event": data})

        await self._dispatch_voice_update(self._voice_state)

    async def on_voice_state_update(self, data: Dict[str, Any]) -> None:
        self._voice_state.update({"sessionId": data["session_id"]})

        channel_id = data["channel_id"]
        if not channel_id:  # We're disconnecting
            self._voice_state.clear()
            return

        self.channel = self.guild.get_channel(int(channel_id))  # type: ignore
        await self._dispatch_voice_update({**self._voice_state, "event": data})

    async def _dispatch_voice_update(self, voice_state: Dict[str, Any]) -> None:
        logger.debug(f"Dispatching voice update:: {self.channel.id}")

        if {"sessionId", "event"} == self._voice_state.keys():
            await self.node._websocket.send(
                op="voiceUpdate", guildId=str(self.guild.id), **voice_state
            )

    async def connect(self, *, timeout: float, reconnect: bool, **kwargs: Any) -> None:
        await self.guild.change_voice_state(channel=self.channel, **kwargs)
        self._connected = True

        logger.info(f"Connected to voice channel:: {self.channel.id}")

    async def disconnect(self, *, force: bool = False) -> None:
        try:
            logger.info(f"Disconnected from voice channel:: {self.channel.id}")

            await self.guild.change_voice_state(channel=None)
            self._connected = False
        finally:
            with contextlib.suppress(ValueError):
                self.node._players.remove(self)

            payload = {"op": "destroy", "guildId": str(self.guild.id)}
            await self.node._websocket.send(**payload)

            self.cleanup()

    async def move_to(self, channel: discord.VoiceChannel) -> None:
        """|coro|

        Moves the player to a different voice channel.

        Parameters
        -----------
        channel: :class:`discord.VoiceChannel`
            The channel to move to. Must be a voice channel.
        """
        await self.guild.change_voice_state(channel=channel)
        logger.info(f"Moving to voice channel:: {channel.id}")

    async def play(
        self,
        source: abc.Playable,
        replace: bool = True,
        start: Optional[int] = None,
        end: Optional[int] = None,
        volume: Optional[int] = None,
        pause: Optional[bool] = None,
    ):
        """|coro|

        Play a WaveLink Track.

        Parameters
        ----------
        source: :class:`abc.Playable`
            The :class:`abc.Playable` track to start playing.
        replace: bool
            Whether this track should replace the current track. Defaults to ``True``.
        start: Optional[int]
            The position to start the track at in milliseconds.
            Defaults to ``None`` which will start the track at the beginning.
        end: Optional[int]
            The position to end the track at in milliseconds.
            Defaults to ``None`` which means it will play until the end.
        volume: Optional[int]
            Sets the volume of the player. Must be between ``0`` and ``1000``.
            Defaults to ``None`` which will not change the volume.
        pause: bool
            Changes the players pause state. Defaults to ``None`` which will not change the pause state.

        Returns
        -------
        :class:`wavelink.abc.Playable`
            The track that is now playing.
        """

        if not replace and self.is_playing():
            return

        await self.update_state({"state": {}})

        if isinstance(source, PartialTrack):
            source = await source._search()

        self._source = source

        payload = {
            "op": "play",
            "guildId": str(self.guild.id),
            "track": source.id,
            "noReplace": not replace,
        }
        if start is not None and start > 0:
            payload["startTime"] = str(start)
        if end is not None and end > 0:
            payload["endTime"] = str(end)
        if volume is not None:
            self._volume = volume
            payload["volume"] = str(volume)
        if pause is not None:
            self._paused = pause
            payload["pause"] = pause

        await self.node._websocket.send(**payload)

        logger.debug(f"Started playing track:: {str(source)} ({self.channel.id})")
        return source

    def is_connected(self) -> bool:
        """Indicates whether the player is connected to voice."""
        return self._connected

    def is_playing(self) -> bool:
        """Indicates whether a track is currently being played."""
        return self.is_connected() and self._source is not None

    def is_paused(self) -> bool:
        """Indicates whether the currently playing track is paused."""
        return self._paused

    async def stop(self) -> None:
        """|coro|

        Stop the Player's currently playing song.
        """
        await self.node._websocket.send(op="stop", guildId=str(self.guild.id))
        logger.debug(f"Current track stopped:: {str(self.source)} ({self.channel.id})")
        self._source = None

    async def set_pause(self, pause: bool) -> None:
        """|coro|

        Set the players paused state.

        Parameters
        ----------
        pause: bool
            A bool indicating if the player's paused state should be set to True or False.
        """
        await self.node._websocket.send(
            op="pause", guildId=str(self.guild.id), pause=pause
        )
        self._paused = pause
        logger.info(f"Set pause:: {self._paused} ({self.channel.id})")

    async def pause(self) -> None:
        """|coro|

        Pauses the player if it was playing.
        """
        await self.set_pause(True)

    async def resume(self) -> None:
        """|coro|

        Resumes the player if it was paused.
        """
        await self.set_pause(False)

    async def set_volume(self, volume: int) -> None:
        """|coro|

        Sets the player's volume. Accepts an int between ``0`` and ``1000`` where ``100`` means 100% volume.

        Parameters
        ----------
        volume: int
            The volume to set the player to.
        """

        self._volume = max(min(volume, 1000), 0)

        await self.node._websocket.send(
            op="volume",
            guildId=str(self.guild.id),
            volume=self.volume
        )
        logger.debug(f"Set volume:: {self.volume} ({self.channel.id})")

    async def seek(self, position: int = 0) -> None:
        """|coro|

        Seek to the given position in the song.

        Parameters
        ----------
        position: int
            The position as an int in milliseconds to seek to.
        """
        await self.node._websocket.send(
            op="seek", guildId=str(self.guild.id), position=position
        )

    async def set_filter(
        self,
        _filter: Filter,
        /, *,
        seek: bool = False
    ) -> None:
        """|coro|

        Set the player's filter.

        Parameters
        ----------
        filter: :class:`wavelink.Filter`
            The filter to apply to the player.
        seek: bool
            Whether to seek the player to its current position
            which will apply the filter immediately. Defaults to ``False``.
        """

        self._filter = _filter

        await self.node._websocket.send(
            op="filters",
            guildId=str(self.guild.id),
            **_filter._payload
        )
        logger.debug(f"Set filter:: {self._filter} ({self.channel.id})")

        if self.is_playing() and seek:
            await self.seek(int(self.position * 1000))
